package lbms.plugins.mldht.kad;

import java.io.IOException;
import java.io.Serializable;
import java.net.InetSocketAddress;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import lbms.plugins.mldht.kad.DHT.LogLevel;
import lbms.plugins.mldht.kad.messages.MessageBase;
import lbms.plugins.mldht.kad.messages.PingRequest;

/**
 * A KBucket is just a list of KBucketEntry objects.
 *
 * The list is sorted by time last seen :
 * The first element is the least recently seen, the last
 * the most recently seen.
 *
 * @author Damokles
 */
public class KBucket implements RPCCallListener, Serializable {

	private static final class ReplacementEntry {
		public ReplacementEntry(KBucketEntry toReplace, KBucketEntry replacement)
		{
			this.toReplace = toReplace;
			this.replacement = replacement;
		}

		final KBucketEntry toReplace;
		final KBucketEntry replacement;

		@Override
        public boolean equals(Object obj) {
			if(!(obj instanceof ReplacementEntry)) {
	            return false;
            }
			return toReplace.equals(((ReplacementEntry)obj).toReplace);
		}
	}

	// any modifying actions to entries must happen through removeAndInsert (which is synchronized)!
	private LinkedList<KBucketEntry>					entries;
	private LinkedList<KBucketEntry>					pending_entries;
	private transient RPCServerBase						srv;
	private transient Node								node;
	private transient Map<RPCCall, ReplacementEntry>	pending_entries_busy_pinging;
	private long										last_modified;

	private transient Task								refresh_task;

	public KBucket () {
		entries = new LinkedList<KBucketEntry>();
		pending_entries = new LinkedList<KBucketEntry>();
		pending_entries_busy_pinging = new HashMap<RPCCall, ReplacementEntry>();
	}

	public KBucket (RPCServerBase srv, Node node) {
		this();
		last_modified = System.currentTimeMillis();
		this.node = node;
		this.srv = srv;
	}

	/**
	 * Inserts an entry into the bucket.
	 * @param entry The entry to insert
	 */
	public void insert (KBucketEntry entry) {
		removeAndInsert(null, entry);
	}

	private void removeAndInsert(KBucketEntry toRemove, KBucketEntry toInsert)
	{
		KBucketEntry toPing = null;

		synchronized (entries)
		{
			if (toRemove != null) {
				entries.remove(toRemove);
			}
			if (toInsert == null) {
				return;
			}
			int idx = entries.indexOf(toInsert);
			if (idx != -1)
			{
				// If in the list, move it to the end
				KBucketEntry oldEntry = entries.get(idx);
				if (!oldEntry.getAddress().equals(toInsert.getAddress()))
				{
					// node changed address, check if old node is really dead to prevent impersonation
					pingEntry(oldEntry, toInsert);
					return;
				}
				// do not change instances, only positions
				toInsert = oldEntry;
				toInsert.signalHasResponded();
				adjustTimerOnInsert(toInsert);
				entries.remove(toInsert);
				entries.addLast(toInsert);
			} else if (entries.size() < DHTConstants.MAX_ENTRIES_PER_BUCKET)
			{
				// insert if not already in the list and we still have room
				entries.addLast(toInsert);
				adjustTimerOnInsert(toInsert);
			} else if (!replaceBadEntry(toInsert))
			{
				// older entries displace younger ones (this code path is mostly used on changing node IDs)
				KBucketEntry youngest = Collections.max(entries, new Comparator<KBucketEntry>()
				{
					public int compare(KBucketEntry o1, KBucketEntry o2) {
						return (int) (o1.getCreationTime() - o2.getCreationTime());
					}
				});
				if (youngest.getCreationTime() > toInsert.getCreationTime())
				{
					entries.remove(youngest);
					entries.addLast(toInsert);
					adjustTimerOnInsert(toInsert);
					toPing = youngest;
				} else {
					toPing = toInsert;
				}
			}
		}
		// ping outside the synchronized block to avoid deadlocks due to dual lock usage with entries + pending entries
		if(toPing != null) {
			pingQuestionable(toPing);
		}
	}

	/**
	 * Get the least recently seen node.
	 *
	 * @return the Entry that was least recently seen
	 */
	public KBucketEntry getLeastRecentlySeen () {
		return entries.get(0);
	}

	/**
	 * Get the number of entries.
	 *
	 * @return The number of entries in this Bucket
	 */
	public int getNumEntries () {
		return entries.size();
	}

	/**
	 * @return the entries
	 */
	public List<KBucketEntry> getEntries () {
		return new ArrayList<KBucketEntry>(entries);
	}

	/**
	 * Checks if this bucket contains an entry.
	 *
	 * @param entry Entry to check for
	 * @return true if found
	 */
	public boolean contains (KBucketEntry entry) {
		return entries.contains(entry);
	}

	/**
	 * Find the K closest entries to a key and store them in the KClosestNodesSearch
	 * object.
	 * @param kns The object to storre the search results
	 * @return true if at least one Node was inserted successully
	 */
	public boolean findKClosestNodes (KClosestNodesSearch kns) {
		//return true if bucket was empty
		//otherwise Node.findKClosestNodes will break
		boolean successfulInsert = entries.size() == 0;
		for (int i = 0; i < entries.size(); i++) {
			if (!entries.get(i).isBad() && kns.tryInsert(entries.get(i))) {
				successfulInsert = true;
			}
		}
		return successfulInsert;
	}

	/**
	 * A peer failed to respond
	 * @param addr Address of the peer
	 */
	public boolean onTimeout (InetSocketAddress addr) {
		for (int i = 0; i < entries.size(); i++) {
			KBucketEntry e = entries.get(i);
			if (e.getAddress() == addr) {
				e.signalRequestTimeout();
				//only removes the entry if it is bad
				removeEntry(e, false);
				return true;
			}
		}
		return false;
	}

	/**
	 * Check if the bucket needs to be refreshed
	 *
	 * @return true if it needs to be refreshed
	 */
	public boolean needsToBeRefreshed () {
		long now = System.currentTimeMillis();

		return (now - last_modified > DHTConstants.BUCKET_REFRESH_INTERVAL
				&& (refresh_task == null || refresh_task.isFinished()) && entries.size() > 0);
	}

	/**
	 * Resets the last modified for this Bucket
	 */
	public void updateRefreshTimer () {
		last_modified = System.currentTimeMillis();
	}

	private void adjustTimerOnInsert(KBucketEntry entry)
	{
		if(entry.getLastResponded() > last_modified) {
			last_modified = entry.getLastResponded();
		}
	}

	/**
	 * Pings an entry, and replaces it if ping fails.
	 *
	 * @param entry entry to ping
	 * @param replacement entry to replace by
	 * @return
	 */
	private boolean pingEntry(KBucketEntry entry, KBucketEntry replacement)
	{
		ReplacementEntry rEntry = new ReplacementEntry(entry,replacement);
		// don't ping if there already is an outstanding ping
		if(pending_entries_busy_pinging.containsValue(rEntry)) {
	        return false;
        }

		PingRequest p = new PingRequest(node.getOurID());
		p.setDestination(entry.getAddress());
		RPCCall c = srv.doCall(p);
		if (c != null) {
			c.addListener(this);
			// add the pending entry
			pending_entries_busy_pinging.put(c, rEntry);
			return true;
		}
		return false;
	}

	private void pingQuestionable (KBucketEntry replacement_entry) {
		if (pending_entries_busy_pinging.size() >= 2) {
			synchronized (pending_entries) {
				//if it is already inserted remove it and add it to the end
				int idx = pending_entries.indexOf(replacement_entry);
				if (idx != -1) {
					replacement_entry = pending_entries.get(idx);
					pending_entries.remove(idx);
				}
				pending_entries.addLast(replacement_entry); // lets not have to many pending_entries calls going on
				if (pending_entries.size() > DHTConstants.MAX_ENTRIES_PER_BUCKET) {
					//remove the least recently seen one
					pending_entries.removeFirst();
				}
			}
			return;
		}

		// we haven't found any bad ones so try the questionable ones
		synchronized (entries)
		{
			for (KBucketEntry e : entries) {
				if (e.isQuestionable() && pingEntry(e, replacement_entry)) {
					return;
				}
			}
		}



		//save the entry if all are good
		synchronized (pending_entries) {
			//if it is already inserted remove it and add it to the end
			int idx = pending_entries.indexOf(replacement_entry);
			if (idx != -1) {
				replacement_entry = pending_entries.get(idx);
				pending_entries.remove(idx);
			}
			pending_entries.addLast(replacement_entry);
			if (pending_entries.size() > DHTConstants.MAX_ENTRIES_PER_BUCKET) {
				//remove the least recently seen one
				pending_entries.removeFirst();
			}
		}
	}

	/**
	 * Tries to instert entry by replacing a bad entry.
	 *
	 * @param entry Entry to insert
	 * @return true if replace was successful
	 */
	private boolean replaceBadEntry (KBucketEntry entry) {
		for (int i = 0; i < entries.size(); i++) {
			KBucketEntry e = entries.get(i);
			if (e.isBad()) {
				// bad one get rid of it
				removeAndInsert(e, entry);
				return true;
			}
		}
		return false;
	}


	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.RPCCallListener#onResponse(lbms.plugins.mldht.kad.RPCCallBase, lbms.plugins.mldht.kad.messages.MessageBase)
	 */
	public void onResponse (RPCCallBase c, MessageBase rsp) {
		ReplacementEntry rEntry = pending_entries_busy_pinging.get(c);

		if(rEntry == null) {
			return;
		}

		pending_entries_busy_pinging.remove(c); // call is done so erase it

		// IDs are equal, address has changed, but old address is still alive -> potential attack
		if(rEntry.toReplace.getID().equals(rEntry.replacement.getID()) && !rEntry.toReplace.getAddress().equals(rEntry.replacement.getAddress()))
		{
			DHT.log("New node "+rEntry.replacement.getAddress()+" claims same Node ID ("+rEntry.toReplace.getID()+") as "+rEntry.toReplace.getAddress()+" ; node dropped as this might be an impersonation attack", LogLevel.Error);
			return;
		}

		synchronized (entries)
		{
			// check if node changed its ID
			for (KBucketEntry entry : entries)
			{
				if (entry.getAddress().equals(rsp.getOrigin()) && !entry.getID().equals(rsp.getID()))
				{
					removeEntry(entry, true); //remove and replace from replacement bucket
					DHT.log("Node " + entry.getAddress() + " changed ID from " + entry.getID() + " to " + rsp.getID(), LogLevel.Info);
					entry = new KBucketEntry(entry.getAddress(), rsp.getID(), entry.getCreationTime());
					entry.signalHasResponded();
					// insert into appropriate bucket for the new ID
					node.insertEntry(entry);
					return;
				}
			}
		}
		// we have a response so try to find the next bad or questionable node
		// if we do not have room see if we can get rid of some bad peers
		if (!replaceBadEntry(rEntry.replacement)) {
			pingQuestionable(rEntry.replacement);
		}

	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.RPCCallListener#onTimeout(lbms.plugins.mldht.kad.RPCCall)
	 */
	public void onTimeout (RPCCallBase c) {
		ReplacementEntry rEntry = pending_entries_busy_pinging.get(c);

		if(rEntry == null) {
			return;
		}

		pending_entries_busy_pinging.remove(c); // call is done so erase it

		removeAndInsert(rEntry.toReplace, rEntry.replacement);
		if(rEntry.toReplace.getID().equals(rEntry.replacement.getID()) && !rEntry.toReplace.getAddress().equals(rEntry.replacement.getAddress())) {
			DHT.log("Node "+rEntry.toReplace.getID()+" changed address from "+rEntry.toReplace.getAddress()+" to "+rEntry.replacement.getAddress(), LogLevel.Info);
		}




		// see if we can do another pending entry
		synchronized (pending_entries) {
			if (pending_entries_busy_pinging.size() < 2
					&& pending_entries.size() > 0) {
				KBucketEntry pe = pending_entries.removeLast();
				if (!replaceBadEntry(pe)) {
					pingQuestionable(pe);
				}
			}
		}
	}

	/**
	 * @param toRemove Entry to remove, if its bad
	 * @param force if true entry will be removed regardless of its state
	 */
	public void removeEntry(KBucketEntry toRemove, boolean force) {
		if (entries.contains(toRemove) && (force || toRemove.isBad()))
		{
			KBucketEntry replacement = null;
			synchronized (pending_entries)
			{
				if (pending_entries.size() > 0) {
					replacement = pending_entries.removeLast();
				}
			}
			removeAndInsert(toRemove, replacement);
		}
	}

	/**
	 * @param srv
	 *            the srv to set
	 */
	public void setServer (RPCServerBase srv) {
		this.srv = srv;
	}

	/**
	 * @param node the node to set
	 */
	public void setNode (Node node) {
		this.node = node;
	}

	/**
	 * @param refresh_task the refresh_task to set
	 */
	public void setRefreshTask (Task refresh_task) {
		this.refresh_task = refresh_task;
	}

	private void readObject(java.io.ObjectInputStream stream) throws IOException, ClassNotFoundException
	{
		pending_entries_busy_pinging = new HashMap<RPCCall, ReplacementEntry>();
		stream.defaultReadObject();
	}


}
