package lbms.plugins.mldht.kad;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.InetAddress;
import java.util.Comparator;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;

import lbms.plugins.mldht.azureus.MlDHTPlugin;
import lbms.plugins.mldht.kad.messages.MessageBase;

import org.gudy.azureus2.plugins.utils.Utilities;

/**
 * @author Damokles
 *
 */
public class Node {

	private Key our_id;
	private KBucket[] bucket = new KBucket[160];
	private RPCServerBase srv;
	private int num_receives;
	private int num_entries;

	private static final int TABLECACHE_FILE_VERSION_CURRENT = 3;
	private static final int TABLECACHE_FILE_VERSION_IPSTORE = 2;
	private static final int TABLECACHE_FILE_VERSION_EXP2_ESTIMATE = 3;

	/**
	 * @param srv
	 */
	public Node(RPCServerBase srv) {
		this.srv = srv;
		num_receives = 0;
		num_entries = 0;
		our_id = Key.createRandomKey();
	}

	/**
	 * An RPC message was received, the node must now update
	 * the right bucket.
	 * @param dh_table The DHT
	 * @param msg The message
	 */
	void recieved (DHTBase dh_table, MessageBase msg) {
		int bucketID = our_id.findApproxKeyDistance(msg.getID());

		// return if bit_on is not good
		if (bucketID >= 160 || bucketID == 0) {
			return;
		}

		// make the bucket if it doesn't exist
		if (bucket[bucketID] == null) {
			bucket[bucketID] = new KBucket(srv, this);
		}

		// insert it into the bucket
		KBucketEntry newEntry = new KBucketEntry(msg.getOrigin(), msg.getID());
		newEntry.setVersion(msg.getVersion());
		bucket[bucketID].insert(newEntry);
		num_receives++;

		/*
		if (num_receives == DHTConstants.MAX_ENTRIES_PER_BUCKET) {
			// do a node lookup upon our own id
			// when we insert the first entries in the table
			NodeLookup nl = dh_table.findNode(our_id);
			if (nl != null) {
				nl.setInfo("Received 8 Nodes, Searching for ourself now.");
			}
		}
		*/

		num_entries = 0;
		for (int i = 1; i < 160; i++) {
			if (bucket[i] != null) {
				num_entries += bucket[i].getNumEntries();
			}
		}
	}

	public void insertEntry (KBucketEntry entry) {
		int bucketID = our_id.findApproxKeyDistance(entry.getID());
		// return if bit_on is not good
		if (bucketID >= 160 || bucketID == 0) {
			return;
		}

		if (bucket[bucketID] == null) {
			bucket[bucketID] = new KBucket(srv, this);
		}
		bucket[bucketID].insert(entry);
	}

	/**
	 * @return OurID
	 */
	public Key getOurID () {
		return our_id;
	}

	/**
	 * Find the K closest entries to a key and store them in the KClosestNodesSearch
	 * object.
	 * @param kns The object to store the search results
	 */
	public void findKClosestNodes (KClosestNodesSearch kns) {
		// go over all buckets until
		int target = our_id.findApproxKeyDistance(kns.getSearchTarget());
		boolean high = true;
		boolean low = true;

		if (bucket[target] != null) {
			bucket[target].findKClosestNodes(kns);
		}

		for (int i = 1; i < bucket.length && (low || high); i++) {
			if (low) {
				if (target - i > 0) {
					if (bucket[target - i] != null) {
						low = bucket[target - i].findKClosestNodes(kns);
					}
				} else {
					low = false;
				}
			}
			if (high) {
				if (target + i < bucket.length) {
					if (bucket[target + i] != null) {
						high = bucket[target + i].findKClosestNodes(kns);
					}
				} else {
					high = false;
				}
			}
		}
	}

	/**
	 * Increase the failed queries count of the bucket entry we sent the message to
	*/
	void onTimeout (MessageBase msg) {
		for (int i = 1; i < bucket.length; i++) {
			if (bucket[i] != null && bucket[i].onTimeout(msg.getDestination())) {
				return;
			}
		}
	}

	/**
	 * Check if a buckets needs to be refreshed, and refresh if necesarry
	 *
	 * @param dh_table
	 */
	public void refreshBuckets (DHTBase dh_table) {
		for (int i = 1; i < bucket.length; i++) {
			KBucket b = bucket[i];
			// remove bootstrap node
			if (b != null
					&& b.getNumEntries() >= DHTConstants.MAX_ENTRIES_PER_BUCKET) {
				List<KBucketEntry> entries = b.getEntries();
				for (KBucketEntry entry : entries) {
					if (entry.getAddress().getAddress().equals(
							DHTConstants.BOOTSTRAP_NODE_ADDRESS)) {
						b.removeEntry(entry, true);
					}
				}

			}
			if (b != null && b.needsToBeRefreshed()) {
				DHT.logDebug("Refreshing Bucket: " + i);
				// the key needs to be the refreshed
				PingRefreshTask nl = dh_table.refreshBucket(b);
				if (nl != null) {
					b.setRefreshTask(nl);
					nl.setInfo("Refreshing Bucket #" + i);
				}
			}
		}
	}

	/**
	 * Check if a buckets needs to be refreshed, and refresh if necesarry
	 *
	 * @param dh_table
	 */
	public void fillBuckets (DHTBase dh_table) {
		boolean foundFirst = false;
		for (int i = 1; i < bucket.length; i++) {
			KBucket b = bucket[i];
			if (b == null && foundFirst) {
				b = bucket[i] = new KBucket(srv, this);
			}

			if (b != null && b.getNumEntries() > 0) {
				foundFirst = true;
			}
			if(!foundFirst) {
				bucket[i] = null;
			}
		}

		for (int i = bucket.length - 1; i >= 1; i--) {
			KBucket b = bucket[i];
			if (b == null) {
				break;
			}
			if (b.getNumEntries() < DHTConstants.MAX_ENTRIES_PER_BUCKET) {
				DHT.logDebug("Filling Bucket: " + i);
				// the key needs to be the refreshed
				NodeLookup nl = dh_table.fillBucket(our_id.createKeyWithDistance(i), b);
				if (nl != null) {
					b.setRefreshTask(nl);
					nl.setInfo("Filling Bucket #" + i);
				}
			}
		}
	}

	/**
	 * Saves the routing table to a file
	 *
	 * @param file to save to
	 * @throws IOException
	 */
	void saveTable (File file) throws IOException {
		ObjectOutputStream oos = null;
		try {
			oos = new ObjectOutputStream(new FileOutputStream(file));
			oos.writeInt(TABLECACHE_FILE_VERSION_CURRENT);
			oos.writeLong(System.currentTimeMillis());
			oos.writeObject(bucket);
			oos.writeObject(our_id);
			oos.writeDouble(PopulationEstimator.getRawDistanceEstimate());
			Utilities utils = MlDHTPlugin.getSingleton().getPluginInterface()
					.getUtilities();
			oos.writeObject(utils.getPublicAddress(false));
			oos.writeObject(utils.getPublicAddress(true));
		} finally {
			if (oos != null) {
				oos.close();
			}
		}
	}

	/**
	 * Loads the routing table from a file
	 *
	 * @param file
	 * @param runWhenLoaded is executed when all load operations are finished
	 * @throws IOException
	 */
	void loadTable (File file, DHT dht, final Runnable runWhenLoaded) {
		boolean runDeferred = false;

		ObjectInputStream ois = null;
		try {
			if (!file.exists()) {
				return;
			}

			ois = new ObjectInputStream(new FileInputStream(file));

			boolean restoreOldID = true;
			// read objects from stream

			// version can be used in the future to skip deserialising some objects that weren't present in previous file formats
			int version = ois.readInt();
			long timestamp = ois.readLong();
			KBucket[] loadedBuckets = (KBucket[]) ois.readObject();
			Key oldID = (Key) ois.readObject();
			double previousEstimate = ois.readDouble();
			PopulationEstimator.setInitialDistanceEstimate(version < TABLECACHE_FILE_VERSION_EXP2_ESTIMATE ? Math.log(previousEstimate)/Math.log(2) : previousEstimate);
			if (version >= TABLECACHE_FILE_VERSION_IPSTORE) {
				InetAddress v4addr = (InetAddress) ois.readObject();
				InetAddress v6addr = (InetAddress) ois.readObject();

				Utilities utils = MlDHTPlugin.getSingleton()
						.getPluginInterface().getUtilities();
				InetAddress v4new = utils.getPublicAddress(false);
				InetAddress v6new = utils.getPublicAddress(true);

				if (v4addr != null && v4new != null && !v4addr.equals(v4new)) {
					restoreOldID = false;
				}
				if (v6addr != null && v6new != null && !v6addr.equals(v6new)) {
					restoreOldID = false;
				}
			}

			restoreOldID |= MlDHTPlugin.getSingleton().getPluginInterface()
					.getPluginconfig().getPluginBooleanParameter(
							"alwaysRestoreID");

			// integrate loaded objects

			int entriesLoaded = 0;

			if (restoreOldID) {
				for (int i = 1; i < loadedBuckets.length; i++) {
					KBucket b = loadedBuckets[i];
					if (b == null) {
						continue;
					}
					if (b.getNumEntries() == 0 && entriesLoaded == 0) {
						b = loadedBuckets[i] = null;
						continue;
					}
					b.setServer(srv);
					b.setNode(this);
					entriesLoaded += b.getNumEntries();
				}
				bucket = loadedBuckets;
				our_id = oldID;
			} else {
				bucket = new KBucket[160];

				// we want to insert the oldest entries first, this way we'll retain longest-lived entries over restarts
				SortedSet<KBucketEntry> entries = new TreeSet<KBucketEntry>(
						new Comparator<KBucketEntry>() {
							public int compare (KBucketEntry o1, KBucketEntry o2) {
								return (int) (o1.getCreationTime() - o2
										.getCreationTime());
							}
						});

				for (int i = 0; i < loadedBuckets.length; i++) {
					if (loadedBuckets[i] != null) {
						entries.addAll(loadedBuckets[i].getEntries());
					}
				}

				//since our_id might have changed, the entries have to be reinserted
				for (KBucketEntry entry : entries) {
					int bucketID = our_id.findApproxKeyDistance(entry.getID());
					// return if bit_on is not good
					if (bucketID >= 160) {
						return;
					}
					/*
					 * insert entries immediately. we can safely add stale entries as
					 * they'll be timed out immediately by a ping lookup and then
					 * replaced from the replacement bucket
					 */
					if (!entry.isBad()) {
						// make the bucket if it doesn't exist
						if (bucket[bucketID] == null) {
							bucket[bucketID] = new KBucket(srv, this);
						}
						bucket[bucketID].insert(entry);
						entriesLoaded++;
					}
				}

			}

			if (entriesLoaded > 0) {
				runDeferred = true;
				PingRefreshTask prt = dht.refreshBuckets(bucket, true);
				prt.setInfo("Pinging cached entries.");
				TaskListener bootstrapListener = new TaskListener() {
					public void finished (Task t) {
						if (runWhenLoaded != null) {
							runWhenLoaded.run();
						}
					}
				};
				prt.addListener(bootstrapListener);
			}

			DHT.logInfo("Loaded " + entriesLoaded + " from cache. Cache was "
					+ ((System.currentTimeMillis() - timestamp) / (60 * 1000))
					+ "min old. Reusing old id = " + restoreOldID);

			return;
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			if (ois != null) {
				try {
					ois.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}

			if (!runDeferred && runWhenLoaded != null) {
				runWhenLoaded.run();
			}
		}
	}

	/**
	 * Get the number of entries in the routing table
	 *
	 * @return
	 */
	public int getNumEntriesInRoutingTable () {
		return num_entries;
	}

	public KBucket[] getBuckets () {
		return bucket;
	}

}
