package lbms.plugins.mldht.kad;

import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;

import lbms.plugins.mldht.kad.messages.FindNodeRequest;
import lbms.plugins.mldht.kad.messages.FindNodeResponse;
import lbms.plugins.mldht.kad.messages.MessageBase;
import lbms.plugins.mldht.kad.messages.MessageBase.Method;
import lbms.plugins.mldht.kad.messages.MessageBase.Type;

/**
 * @author Damokles
 *
 */
public class NodeLookup extends Task {
	private int						validReponsessSinceLastClosestSetModification;
	SortedSet<Key>			closestSet;
	private Map<MessageBase, Key>	lookupMap;
	private boolean fillWithAllBuckets = false;
	
	NodeLookup (Key node_id, RPCServerBase rpc, Node node, boolean isBootstrap) {
		super(rpc, node);
		fillWithAllBuckets = !isBootstrap;
		this.targetKey = node_id;
		this.closestSet = new TreeSet<Key>(new Comparator<Key>() {
			public int compare (Key o1, Key o2) {
				return targetKey.distance(o1).compareTo(targetKey.distance(o2));
			}
		});
		this.lookupMap = new HashMap<MessageBase, Key>();
	}

	@Override
	synchronized void update () {
		// go over the todo list and send find node calls
		// until we have nothing left
		synchronized (todo) {

			while (todo.size() > 0 && canDoRequest()) {
				KBucketEntry e = todo.first();
				todo.remove(e);
				// only send a findNode if we haven't allready visited the node
				if (!visited.contains(e)) {
					// send a findNode to the node
					FindNodeRequest fnr = new FindNodeRequest(node.getOurID(),
							targetKey);
					fnr.setDestination(e.getAddress());
					synchronized (lookupMap) {
						lookupMap.put(fnr, e.getID());
					}
					rpcCall(fnr);
					synchronized (visited) {
						visited.add(e);
					}
				}
				// remove the entry from the todo list
			}
		}

		if (todo.size() == 0 && getNumOutstandingRequests() == 0
				&& !isFinished()) {
			done();
		} else if (validReponsessSinceLastClosestSetModification >= DHTConstants.MAX_CONCURRENT_REQUESTS) {
			done(); // quit after 15 nodes responses
		}
	}

	@Override
	void callFinished (RPCCallBase c, MessageBase rsp) {
		if (isFinished()) {
			return;
		}

		// check the response and see if it is a good one
		if (rsp.getMethod() == Method.FIND_NODE
				&& rsp.getType() == Type.RSP_MSG) {

			MessageBase b = c.getRequest();
			synchronized (lookupMap) {
				if (lookupMap.containsKey(b)) {
					synchronized (closestSet) {
						Key toAdd = lookupMap.remove(b);
						closestSet.add(toAdd);
						if (closestSet.size() > DHTConstants.MAX_ENTRIES_PER_BUCKET) {
							Key last = closestSet.last();
							closestSet.remove(last);
							if (toAdd == last) {
								validReponsessSinceLastClosestSetModification++;
							} else {
								validReponsessSinceLastClosestSetModification = 0;
							}
						}
					}
				}
			}

			FindNodeResponse fnr = (FindNodeResponse) rsp;
			byte[] nodes = fnr.getNodes();
			int nnodes = nodes.length / 26;
			for (int j = 0; j < nnodes; j++) {
				// unpack an entry and add it to the todo list
				KBucketEntry e = PackUtil.UnpackBucketEntry(nodes, j * 26);
				// lets not talk to ourself
				if (!e.getID().equals(node.getOurID()) && !todo.contains(e)
						&& !visited.contains(e)) {
					synchronized (todo) {
						todo.add(e);
					}
				}
			}
		}
	}

	@Override
	void callTimeout (RPCCallBase c) {

	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.Task#start()
	 */
	@Override
	void start () {
		int added = 0;

		// delay the filling of the todo list until we actually start the task
		
		KBucket[] buckets = node.getBuckets();
		outer: for (int i = buckets.length -1; i >= 1; i--)
			if (buckets[i] != null)
				for (KBucketEntry e : buckets[i].getEntries())
					if (!e.isBad())
					{
						todo.add(e);
						added++;
						if (!fillWithAllBuckets && added >= 2 * DHTConstants.MAX_ENTRIES_PER_BUCKET)
							break outer;
					}

		super.start();
	}

	@Override
	protected void done () {
		super.done();

		PopulationEstimator.update(new TreeSet<Key>(closestSet));
	}
}
