package lbms.plugins.mldht.kad;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetSocketAddress;
import java.net.SocketException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.ConcurrentMap;

import lbms.plugins.mldht.kad.DHT.LogLevel;
import lbms.plugins.mldht.kad.messages.MessageBase;
import lbms.plugins.mldht.kad.messages.MessageDecoder;
import lbms.plugins.mldht.kad.messages.PingRequest;
import lbms.plugins.mldht.kad.messages.MessageBase.Type;

import org.gudy.azureus2.core3.util.BDecoder;

/**
 * @author Damokles
 *
 */
public class RPCServer implements Runnable, RPCServerBase {
	
	private static final class ByteWrapper {
		private final byte[] arr;
		private final int hash;
		
		ByteWrapper(byte[] a)
		{
			arr = a;
			hash = Arrays.hashCode(a);
		}
		
		public ByteWrapper(short s)
		{
			this(new byte[] {(byte)(s<<8),(byte)(s&0xff)});
		}
		
		@Override
		public int hashCode() {
			return hash;
		}
		
		@Override
		public boolean equals(Object obj) {
			return obj instanceof ByteWrapper && Arrays.equals(arr,((ByteWrapper)obj).arr);
		}
	}

	private DatagramSocket							sock;
	private DHT										dh_table;
	private ConcurrentMap<ByteWrapper, RPCCallBase>	calls;
	private Queue<RPCCallBase>						call_queue;
	private short									next_mtid;
	private volatile boolean						running;
	private Thread									thread;
	private int										numReceived;
	private int										numSent;
	private int										port;
	private RPCStats								stats;

	public RPCServer (DHT dh_table, int port) throws SocketException { //, Object parent
		this.port = port;
		sock = new DatagramSocket(port);
		calls = new ConcurrentHashMap<ByteWrapper, RPCCallBase>(80,0.75f,3);
		call_queue = new ConcurrentLinkedQueue<RPCCallBase>();
		stats = new RPCStats();
		this.dh_table = dh_table;
	}

	/* (non-Javadoc)
	 * @see java.lang.Runnable#run()
	 */
	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.RPCServerBase#run()
	 */
	public void run() {
		outer: while (running)
		{
			byte[] buffer = new byte[DHTConstants.RECEIVE_BUFFER_SIZE];
			DatagramPacket packet = new DatagramPacket(buffer, buffer.length);
			try
			{
				sock.receive(packet);
				handlePacket(packet);
			} catch (Exception e)
			{
				if (running)
				{
					DHT.log(e, LogLevel.Error);
					while (running && sock.isClosed())
					{
						try
						{  // wait a bit before we restore the socket to avoid screaming loops
							Thread.sleep(200);
							sock = new DatagramSocket(port);
						} catch (Exception e1)
						{
							DHT.log("Reestablishing listening port failed, will retry in 5s", LogLevel.Error);
							DHT.log(e1, LogLevel.Error);
							try
							{  // the old, closed socket might not have been cleaned up properly, wait longer
								Thread.sleep(5000);
							} catch (InterruptedException e2)
							{
								break outer;
							}
						}
					}
				}
			}
		}
		DHT.logInfo("Stopped RPC Server");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see lbms.plugins.mldht.kad.RPCServerBase#start()
	 */
	public void start () {
		DHT.logInfo("Starting RPC Server");
		running = true;
		thread = new Thread(this, "mlDHT RPC Thread");
		thread.setPriority(Thread.MIN_PRIORITY);
		thread.setDaemon(true);
		thread.start();
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.RPCServerBase#stop()
	 */
	public void stop () {
		running = false;
		sock.close();
		thread = null;
		DHT.logInfo("Stopping RPC Server");
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.RPCServerBase#doCall(lbms.plugins.mldht.kad.messages.MessageBase)
	 */
	public RPCCall doCall (MessageBase msg) {
		
		RPCCall c = new RPCCall(this, msg);
		
		short mtid = next_mtid++;
		short start = mtid;
		while (calls.putIfAbsent(new ByteWrapper(mtid),c) != null) {
			mtid = next_mtid++;
			if (next_mtid == start) // if this happens we cannot do any calls
			{
				call_queue.add(c);
				System.out.println("Queueing RPC call, no slots available at the moment");
				return c;
			}
		}
			
		msg.setMTID(mtid);
		sendMessage(msg);
		c.start();

		return c;
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.RPCServerBase#timedOut(byte)
	 */
	public void timedOut (byte[] mtid) {
		// delete the call
		
		ByteWrapper w = new ByteWrapper(mtid);
		RPCCallBase c = calls.get(w);
		if (c != null)
		{
			stats.addTimeoutMessageToCount(c.getRequest());
			calls.remove(w);
			dh_table.timeout(c.getRequest());
		}
			
		doQueuedCalls();
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.RPCServerBase#ping(lbms.plugins.mldht.kad.Key, java.net.InetSocketAddress)
	 */
	public void ping (Key our_id, InetSocketAddress addr) {
		PingRequest pr = new PingRequest(our_id);
		pr.setOrigin(addr);
		doCall(pr);
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.RPCServerBase#findCall(byte)
	 */
	public RPCCallBase findCall (byte[] mtid) {
		return calls.get(new ByteWrapper(mtid));
	}

	/// Get the number of active calls
	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.RPCServerBase#getNumActiveRPCCalls()
	 */
	public int getNumActiveRPCCalls () {
		return calls.size();
	}

	/**
	 * @return the numReceived
	 */
	public int getNumReceived () {
		return numReceived;
	}

	/**
	 * @return the numSent
	 */
	public int getNumSent () {
		return numSent;
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.RPCServerBase#getStats()
	 */
	public RPCStats getStats () {
		return stats;
	}

	private void handlePacket (DatagramPacket p) {
		numReceived++;
		stats.addReceivedBytes(p.getLength());
		byte[] ip = p.getAddress().getAddress();
		if (DHT.isLogLevelEnabled(LogLevel.Verbose)) {
			try {
				DHT.logVerbose(new String(p.getData(), 0, p.getLength(),
						"UTF-8"));
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		try {
			byte[] buf = new byte[p.getLength()];
			System.arraycopy(p.getData(), p.getOffset(), buf, 0, buf.length);
			Map<String, Object> bedata = BDecoder.decode(buf);
			MessageBase msg = MessageDecoder.parseMessage(bedata, this);
			if (msg != null) {
				DHT.logDebug("RPC received message ["+(ip[0] & 0xFF) + "."
					+ (ip[1] & 0xFF) + "." + (ip[2] & 0xFF) + "." + (ip[3] & 0xFF)+"] "+msg.toString());
				stats.addReceivedMessageToCount(msg);
				msg.setOrigin(new InetSocketAddress(p.getAddress(), p.getPort()));
				msg.apply(dh_table);
				// erase an existing call
				if (msg.getType() == Type.RSP_MSG
						&& calls.containsKey(new ByteWrapper(msg.getMTID()))) {
					RPCCallBase c = calls.get(new ByteWrapper(msg.getMTID()));
					if(c.getRequest().getDestination().equals(msg.getOrigin()))
					{
						// delete the call, but first notify it off the response
						c.response(msg);
						calls.remove(new ByteWrapper(msg.getMTID()));
						doQueuedCalls();						
					} else
						DHT.logInfo("Response source ("+msg.getOrigin()+") mismatches request destination ("+c.getRequest().getDestination()+"); ignoring response");
				}
			} else
			{
				try {
					DHT.logDebug("RPC received message [" + (ip[0] & 0xFF) + "."
						+ (ip[1] & 0xFF) + "." + (ip[2] & 0xFF) + "." + (ip[3] & 0xFF) + "] Decode failed msg was:"+new String(p.getData(), 0, p.getLength(),"UTF-8"));
				} catch (Exception e) {
					e.printStackTrace();
				}
			}

				
		} catch (IOException e) {
			DHT.log(e, LogLevel.Debug);
		}
	}

	/* (non-Javadoc)
	 * @see lbms.plugins.mldht.kad.RPCServerBase#sendMessage(lbms.plugins.mldht.kad.messages.MessageBase)
	 */
	public void sendMessage (MessageBase msg) {
		try {
			stats.addSentMessageToCount(msg);
			send(msg.getDestination(), msg.encode());
			byte[] ip = msg.getDestination().getAddress().getAddress();
			DHT.logDebug("RPC send Message: [" + (ip[0] & 0xFF) + "."	+ (ip[1] & 0xFF) + "." + (ip[2] & 0xFF) + "." + (ip[3] & 0xFF) + "] "+ msg.toString());
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void send (InetSocketAddress addr, byte[] msg) throws IOException {
		if (!sock.isClosed()) {
			DatagramPacket p = new DatagramPacket(msg, msg.length);
			p.setAddress(addr.getAddress());
			p.setPort(addr.getPort());
			sock.send(p);
			stats.addSentBytes(msg.length);
			numSent++;
		}
	}

	private void doQueuedCalls () {
		while (call_queue.peek() != null && calls.size() < DHTConstants.MAX_ACTIVE_CALLS) {
			RPCCallBase c;
			MessageBase msg;

			if((c = call_queue.poll()) == null)
				return;

			
			short mtid = next_mtid++;
			short start = mtid;
			boolean canStart = true;
			while (calls.putIfAbsent(new ByteWrapper(mtid), c) != null) {
				mtid = next_mtid++;
				if(next_mtid == start)
				{
					canStart = false;
					break;
				}
			}
			
			if(!canStart)
			{
				call_queue.add(c);
				return;
			}
				

			msg = c.getRequest();
			msg.setMTID(mtid);
			sendMessage(msg);
			c.start();
		}
	}

}
