package lbms.plugins.mldht.kad.messages;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import lbms.plugins.mldht.kad.DBItem;
import lbms.plugins.mldht.kad.DHT;
import lbms.plugins.mldht.kad.DHTConstants;
import lbms.plugins.mldht.kad.Key;
import lbms.plugins.mldht.kad.PackUtil;
import lbms.plugins.mldht.kad.RPCCallBase;
import lbms.plugins.mldht.kad.RPCServerBase;
import lbms.plugins.mldht.kad.DHT.LogLevel;
import lbms.plugins.mldht.kad.messages.MessageBase.Method;

/**
 * @author Damokles
 *
 */
public class MessageDecoder {

	public static MessageBase parseMessage (Map<String, Object> map,
			RPCServerBase srv) {

		try {
			String vn = getStringFromBytes((byte[]) map.get(DHTConstants.TYP));
			if (vn == null) {
				return null;
			}

			String version = getStringFromBytes((byte[]) map.get(DHTConstants.VER));

			MessageBase mb = null;
			if (vn.equals(DHTConstants.REQ)) {
				mb = parseRequest(map);
			} else if (vn.equals(DHTConstants.RSP)) {
				mb = parseResponse(map, srv);
			} else if (vn.equals(DHTConstants.ERR)) {
				mb = parseError(map);
			}

			if (mb != null && version != null) {
				mb.setVersion(version);
			}

			return mb;
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * @param map
	 * @return
	 */
	private static MessageBase parseError (Map<String, Object> map) {
		List<Object> errmap = (List<Object>) map.get(DHTConstants.ERR);
		if (errmap == null || !map.containsKey(DHTConstants.TID)) {
			return null;
		}

		byte[] mtid;
		if (map.get(DHTConstants.TID) instanceof Long) {
			return null;//mtid = ((Long) map.get(DHTConstants.TID)).byteValue();
		} else if (map.get(DHTConstants.TID) instanceof byte[]) {
			byte[] ba = (byte[]) map.get(DHTConstants.TID);

			if (ba == null || ba.length < 1) {
				return null;
			}
			mtid = ba;
		} else {
			return null;
		}

		return new ErrorMessage(mtid, ((Long) errmap.get(0)).intValue(),
				getStringFromBytes((byte[]) errmap.get(1)));
	}

	/**
	 * @param map
	 * @param srv
	 * @return
	 */
	private static MessageBase parseResponse (Map<String, Object> map,
			RPCServerBase srv) {
		Map<String, Object> args = (Map<String, Object>) map.get(DHTConstants.RSP);
		if (args == null || map.get(DHTConstants.TID) == null) {
			DHT.logDebug("ParseRsp : args || !args.getValue(id) || !dict.getValue(TID)");
			return null;
		}

		byte[] mtid;
		if (map.get(DHTConstants.TID) instanceof Long) {
			return null;//mtid = ((Long) map.get(DHTConstants.TID)).byteValue();
		} else if (map.get(DHTConstants.TID) instanceof byte[]) {
			byte[] ba = (byte[]) map.get(DHTConstants.TID);

			if (ba == null || ba.length != 2) {
				return null;
			}
			mtid = ba;
		} else {
			return null;
		}

		// find the call
		RPCCallBase c = srv.findCall(mtid);
		if (c == null) {
			DHT.logDebug("Cannot find RPC call for response: " + new String(mtid));
			return null;
		}

		return parseResponse(map, c.getMessageMethod(), mtid);
	}

	/**
	 * @param map
	 * @param msgMethod
	 * @param mtid
	 * @return
	 */
	private static MessageBase parseResponse (Map<String, Object> map,
			Method msgMethod, byte[] mtid) {
		Map<String, Object> args = (Map<String, Object>) map.get(DHTConstants.RSP);
		if (args == null || !args.containsKey("id")) {
			return null;
		}

		byte[] hash = (byte[]) args.get("id");

		if (hash == null || hash.length != Key.SHA1_HASH_LENGTH) {
			return null;
		}

		Key id = new Key(hash);

		switch (msgMethod) {
		case PING:
			return new PingResponse(mtid, id);
		case FIND_NODE:
			if (!args.containsKey("nodes")) {
				return null;
			} else {
				return new FindNodeResponse(mtid, id,
						(byte[]) args.get("nodes"));
			}
		case GET_PEERS:
			if (args.containsKey("token")) {
				hash = (byte[]) args.get("token");

				if (hash == null || hash.length != Key.SHA1_HASH_LENGTH) {
					return null;
				}

				Key token = new Key(hash);
				byte[] data;
				List<Object> vals = (List<Object>) args.get("values");

				if (vals != null && vals.size() == 1) {
					List<DBItem> dbl = Arrays.asList(PackUtil.UnpackDBItems((byte[])vals.get(0)));
					return new GetPeersResponse(mtid, id, dbl, token);
				} else if (vals != null && vals.size() > 1) {
					List<DBItem> dbl = new ArrayList<DBItem>(vals.size());
					for (int i=0;i<vals.size();i++) {
						if (((byte[])vals.get(i)).length < 6) {
	                        continue;
                        }
						dbl.add(new DBItem((byte[])vals.get(i)));
					}
					return new GetPeersResponse(mtid, id, dbl, token);
				} else if (args.containsKey("nodes")) {
					data = (byte[]) args.get("nodes");
					return new GetPeersResponse(mtid, id, data, token);
				} else {
					DHT.logDebug("No nodes or values in get_peers response");
					return null;
				}
			} else {
				DHT.logDebug("No token in get_peers response");
			}
		case ANNOUNCE_PEER:
			return new AnnounceResponse(mtid, id);
		default:
			return null;
		}
	}

	/**
	 * @param map
	 * @return
	 */
	private static MessageBase parseRequest (Map<String, Object> map) {
		Object vn = map.get(DHTConstants.REQ);
		Map<String, Object> args = (Map<String, Object>) map.get(DHTConstants.ARG);
		if (vn == null || args == null) {
			return null;
		}

		if (!args.containsKey("id")) {
			return null;
		}

		if (!map.containsKey(DHTConstants.TID)) {
			return null;
		}

		byte[] hash = (byte[]) args.get("id");

		if (hash == null || hash.length != Key.SHA1_HASH_LENGTH) {
			return null;
		}

		Key id = new Key(hash);

		byte[] mtid;
		if (map.get(DHTConstants.TID) instanceof Long) {
			return null;//mtid = ((Long) map.get(DHTConstants.TID));
		} else if (map.get(DHTConstants.TID) instanceof byte[]) {
			byte[] ba = (byte[]) map.get(DHTConstants.TID);

			if (ba == null || ba.length < 1) {
				return null;
			}
			mtid = ba;
		} else {
			return null;
		}

		MessageBase msg = null;

		String str = getStringFromBytes((byte[]) vn);
		if (str.equalsIgnoreCase("ping")) {
			msg = new PingRequest(id);
		} else if (str.equalsIgnoreCase("find_node")) {
			if (args.containsKey("target")) {
				hash = (byte[]) args.get("target");
				if (hash == null || hash.length != Key.SHA1_HASH_LENGTH) {
					return null;
				}
				msg = new FindNodeRequest(id, new Key(hash));

			}
		} else if (str.equalsIgnoreCase("get_peers")) {
			if (args.containsKey("info_hash")) {
				hash = (byte[]) args.get("info_hash");
				if (hash == null || hash.length != Key.SHA1_HASH_LENGTH) {
					return null;
				}
				msg = new GetPeersRequest(id, new Key(hash));
			}
		} else if (str.equalsIgnoreCase("announce_peer")) {
			if (args.containsKey("info_hash") && args.containsKey("port")
					&& args.containsKey("token")) {
				hash = (byte[]) args.get("info_hash");
				if (hash == null || hash.length != Key.SHA1_HASH_LENGTH) {
					return null;
				}
				Key infoHash = new Key(hash);

				hash = (byte[]) args.get("token");
				if (hash == null || hash.length != Key.SHA1_HASH_LENGTH) {
					return null;
				}

				msg = new AnnounceRequest(id, infoHash,
						((Long) args.get("port")).intValue(), new Key(hash));
			}
		} else {
			DHT.logDebug("Received unknown Message Type: "+str);
		}

		if (msg != null) {
			msg.setMTID(mtid);
		}

		return msg;
	}

	private static String getStringFromBytes (byte[] bytes) {
		if (bytes == null) {
			return null;
		}
		try {
			return new String(bytes, "UTF-8");
		} catch (Exception e) {
			DHT.log(e, LogLevel.Verbose);
			return null;
		}
	}
}
